/*
 *  ArgumentParser class
 *  Copyright (c) 2003 Jachym Kouba (i-jkouba)
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *	$Id: ArgumentParser.cs, v1.3 2004/02/17 03:01:21 jachymko $
 * 
 */

namespace Argument {
	using System;
	using System.Collections;
	using System.Collections.Specialized;

	/// <summary>
	/// Main argument-parsing class
	/// </summary>
	public class Parser {
		private FlagCollection _flags = new FlagCollection();
		private bool _moreArgsAllowed = true;
		private int _minRequiredArgCount = 0;
		private string _programDesc = string.Empty;

		private static bool useHelp = true;

		/// <summary>
		/// Will be shown when argument count &lt; MinRequiredArgCount
		/// </summary>
		public string UsageString = "{1}\r\nUsage: {0} [OPTION]...\r\n\r\nTry \'{0} --help\' for more options.";

		/// <summary>
		/// Gets or sets value indicating wheter internal help generator should be used
		/// </summary>
		public static bool UseGeneratedHelp {
			get {
				return useHelp;
			}
			set {
				useHelp = value;
			}
		}

		/// <summary>
		/// A short description of the program, will be displayed in help and usage
		/// </summary>
		public string ProgramDescription {
			get {
				return _programDesc;
			}
			set {
				_programDesc = value;
			}
		}
		/// <summary>
		/// Specifies whether to allow unflagged arguments - like URLs, file names, etc. They must
		/// after the last regular flagged argument
		/// </summary>
		public bool MoreArgsAllowed {
			get {
				return _moreArgsAllowed;
			}
			set {
				_moreArgsAllowed = value;
			}
		}
		/// <summary>
		/// Specifies the minimum argument count not to show the usage message. (if you require at least one URL,
		/// leave it one. If you can work w/o parameters, set to zero 
		/// </summary>
		public int MinRequiredArgCount {
			get {
				return _minRequiredArgCount;
			}
			set {
				_minRequiredArgCount = value;
			}
		}
		/// <summary>
		/// Collection of all used flags
		/// </summary>
		public FlagCollection Flags {
			get {
				return _flags;
			}
		}
		/// <summary>
		/// Parses the custom argument-array
		/// </summary>
		/// <param name="arguments">Custom array of arguments</param>
		/// <returns>The remaining unprocessed arguments, if they're allowed. <seealso cref="MoreArgsAllowed" /></returns>
		public string[] Parse(string[] arguments) {
			try {
				string[] extra;
				InnerParse(arguments, out extra);

				return extra;
			}
			catch(Exception exc) {
				Console.WriteLine(exc.Message);
				return new string[0];
			}
		}

		public bool Parse(string[] arguments, out string[] extra) {
			try {
				return InnerParse(arguments, out extra);
			} catch(Exception exc) {
				Console.WriteLine(exc.Message);
				extra = new string[0];
				return false;
			}
		}

		public void DisplayUsage() {
			Console.WriteLine(UsageString, System.IO.Path.GetFileNameWithoutExtension(System.Diagnostics.Process.GetCurrentProcess().MainModule.FileName), _programDesc);
		}

		/// <summary>
		/// Parses the current arguments
		/// </summary>
		/// <returns>The remaining unprocessed arguments, if they're allowed. <seealso cref="MoreArgsAllowed" /></returns>
		public string[] Parse() {
			// remove the executable name
			ArrayList arguments = new ArrayList(Environment.GetCommandLineArgs());
			arguments.RemoveAt(0);

			return Parse((string[])arguments.ToArray(typeof(string)));
		}
		private bool InnerParse(string[] arguments, out string[] extra) {
			if(arguments.Length < _minRequiredArgCount) {
				DisplayUsage();
				extra  = new string[0];
				return false;
			}

			ArrayList args = new ArrayList(arguments);

			for(int i = 0; i < args.Count; i++) {
				string arg = ((string)args[i]).Trim();
				Flag flag = null;
				bool longForm = false;

				// if the first char is "-" and second is something else, it's the short form
				if(arg[0]=='-' && arg[1]!='-') {
					// remove the "-"
					arg = arg.Substring(1);
					flag = _flags.GetByShortForm(arg);
					longForm = false;
				}
					// if the arg begins with '--', it's the long form
				else if(arg[0]=='-' && arg[1]=='-') {
					// remove the "--"
					arg = arg.Substring(2);
					flag = _flags.GetByLongForm(arg);
					longForm = true;
				}
					// if it begins with something else, it isn't a flag but an uncatched value,
					// so we stop enumerating the StringCollection and return the remaining strings
					// as a simple array
				else
					break;

				// Show the help
				if(((longForm && arg==Flag.HelpString) || (!longForm && arg==Flag.QuestionMark)) && useHelp) {
					// write out the description
					Console.WriteLine(_programDesc + "\n");

					foreach(Flag f in _flags) {
						// clear all flags
						f.Set(false);

						if(f.ShortForm.Length == 0)
							if(f.HasValue) {
								string val = f.ValueSample.Length > 0 ? f.ValueSample : "VALUE";
								Console.WriteLine("--{0}={1}\r\n\t{2}\r\n", f.LongForm, val, f.Description);
							} else {
								Console.WriteLine("--{0}\r\n\t{1}\r\n", f.LongForm, f.Description);
							}
						else {
							if(f.HasValue) {
								string val = f.ValueSample.Length > 0 ? f.ValueSample : "VALUE";
								Console.WriteLine("-{0} {2}; --{1}={2}\r\n\t{3}\r\n", f.ShortForm, f.LongForm, f.ValueSample, f.Description);
							} else {
								Console.WriteLine("-{0}; --{1}\r\n\t{2}\r\n", f.ShortForm, f.LongForm, f.Description);
							}
						}
					}

					args.Clear();
					extra = new string[0];
					return false;
				}

				// the flag wasn't found; it is an unknown argument
				if(flag==null)
					throw(new Exceptions.UnknownArgument(arg)); // unknown argument

				// set the flag
				flag.Set(true);

				// look for the value (short form version) and remove it from the StringCollection
				if(flag.HasValue && !longForm) {
					if(i+1==args.Count)
						throw(new Exceptions.MissingArgumentValue(arg)); // missing argument value

					string val = (string)args[i+1];

					if(val.StartsWith("-"))
						throw(new Exceptions.MissingArgumentValue(arg)); // next string is an agrument, too

					args.RemoveAt(i+1);
					flag.SetValue(val);
				}
					// look for the value (long form version)
				else if(flag.HasValue && longForm && arg.IndexOf('=')!=-1) {
					string val = arg.Substring(arg.IndexOf('=')+1);
					arg = arg.Substring(0, arg.IndexOf('='));

					if(val.Trim()=="") throw(new Exceptions.MissingArgumentValue(arg));
					flag.SetValue(val);
				}
				
				// remove the argument from the collection, as it is processed
				args[i] = string.Empty;
			}

			// If allowed, return the remaining un-flagged arguments
			if(args.Count > 0 && _moreArgsAllowed) {
				for(int j = args.Count-1; j >= 0; j--)
					if((string)args[j]==string.Empty)
						args.RemoveAt(j);

				extra = (string[])args.ToArray(typeof(string));
				return true;
			}
				// if not allowed, throw an exception
			else if(args.Count > 0 && !_moreArgsAllowed)
				throw(new Exceptions.UnknownArgument((string)args[0]));
			else {
				extra = new string[0];
				return true;
			}
		}
	}
	/// <summary>
	/// Represents one flag
	/// </summary>
	public class Flag {
		private string _key = string.Empty;
		private string _shortForm = string.Empty;
		private string _longForm = string.Empty;
		private bool _hasValue = false;
		private string _description = string.Empty;
		private string _value = string.Empty;
		private string _valueSample = string.Empty;
		private bool _set = false;

		public static readonly string HelpString = "help";
		public static readonly string QuestionMark = "?";

		/// <summary>
		/// Constructs new instance of Flag class using all parameters
		/// </summary>
		/// <param name="key">Unique key of this flag</param>
		/// <param name="shortForm">Short form of argument specifying this flag</param>
		/// <param name="longForm">Expanded form of argument specifying this flag</param>
		/// <param name="hasValue">Specifies whether argument following this one is its value or not</param>
		/// <param name="description">Description which will be shown with in the help</param>
		public Flag(string key, string longForm, string shortForm, bool hasValue, string valueSample, string description) {
			// Validate the parameters
			if(key==null||key==string.Empty)
				throw(new ArgumentNullException("key"));
			if(shortForm==null)
				throw(new ArgumentNullException("shortForm"));
			if(longForm==null)
				throw(new ArgumentNullException("longForm"));
			if(description==null)
				throw(new ArgumentNullException("description"));
			if((longForm==HelpString || shortForm==QuestionMark) && Parser.UseGeneratedHelp )
				throw(new Exceptions.HelpFlagReserved()); // --help and -? arguments are reserved for us
			
			// Set the variables
			_key = key;
			_shortForm = shortForm;
			_longForm = longForm;
			_hasValue = hasValue;
			_description = description;
			_valueSample = valueSample;
		}

		/// <summary>
		/// Unique key of this flag
		/// </summary>
		public string Key {
			get {
				return _key;
			}
		}
		/// <summary>
		/// The short form of argument specifying this flag (ex. I)
		/// </summary>
		public string ShortForm {
			get {
				return _shortForm;
			}
		}
		/// <summary>
		/// The expanded form of argument specifying this flag (ex. include-directories)
		/// </summary>
		public string LongForm {
			get {
				return _longForm;
			}
		}
		/// <summary>
		/// True if this flag requires an value, otherwise it is a simple boolean
		/// </summary>
		public bool HasValue {
			get {
				return _hasValue;
			}
		}
		/// <summary>
		/// Description of this flag in automatically generated help
		/// </summary>
		public string Description {
			get {
				return _description;
			}
		}
		/// <summary>
		/// Returns the user-supplied value of this flag, if applicable
		/// </summary>
		public string Value {
			get {
				if(!_hasValue)
					throw(new InvalidOperationException("Cannot query the value when HasValue==false"));

				return _value.Trim();
			}
		}
		/// <summary>
		/// Returns whether the flag was set or not
		/// </summary>
		public bool IsSet {
			get {
				if(_set && !_hasValue)
					return true;
				else if(_set && _hasValue && Value!=string.Empty)
					return true;
				else
					return false;
			}
		}

		/// <summary>
		/// Returns sample value of this flag (used in help)
		/// </summary>
		public string ValueSample {
			get {
				return _valueSample;
			}
		}

	
		internal void Set(bool value) {
			_set = value;
		}
		internal void SetValue(string val) {
			_value = val;
		}
	}

	/// <summary>
	/// Collection of Flag class' instances
	/// </summary>
	public class FlagCollection : IEnumerable {
		private Hashtable _innerHashtable = new Hashtable();

		IEnumerator IEnumerable.GetEnumerator() {
			return _innerHashtable.Values.GetEnumerator();
		}
		/// <summary>
		/// Returns count of Flags in the collection
		/// </summary>
		public int Count {
			get {
				return _innerHashtable.Count;
			}
		}
		/// <summary>
		/// Adds new Flag into the collection
		/// </summary>
		/// <param name="flag">Flag to be added</param>
		public void Add(Flag flag) {
			_innerHashtable.Add(flag.Key, flag);
		}
		/// <summary>
		/// Adds new Flag into the collection
		/// </summary>
		/// <param name="key">Unique key of this flag</param>
		/// <param name="longForm">The expanded form of argument specifying this flag (ex. include-directories)</param>
		/// <param name="shortForm">The short form of argument specifying this flag (ex. I)</param>
		/// <param name="hasValue">True if this flag requires an value, otherwise it is a simple boolean</param>
		/// <param name="sampleValue">Sample value of this flag (for usage help)</param>
		/// <param name="description">Description of this flag in automatically generated usage help</param>
		public void Add(string key, string longForm, string shortForm, bool hasValue, string sampleValue, string description) {
			Add(new Flag(key, longForm, shortForm, hasValue, sampleValue, description));
		}
		/// <summary>
		/// Adds new Flag into the collection
		/// </summary>
		/// <param name="key">Unique key of this flag</param>
		/// <param name="longForm">The expanded form of argument specifying this flag (ex. include-directories)</param>
		public void Add(string key, string longForm) {
			Add(key, longForm, string.Empty, false, string.Empty, string.Empty);
		}
		/// <summary>
		/// Adds new Flag into the collection
		/// </summary>
		/// <param name="key">Unique key of this flag</param>
		/// <param name="longForm">The expanded form of argument specifying this flag (ex. include-directories)</param>
		/// <param name="shortForm">The short form of argument specifying this flag (ex. I)</param>
		public void Add(string key, string longForm, string shortForm) {
			Add(key, longForm, shortForm, false, string.Empty, string.Empty);
		}
		/// <summary>
		/// Adds new Flag into the collection
		/// </summary>
		/// <param name="key">Unique key of this flag</param>
		/// <param name="longForm">The expanded form of argument specifying this flag (ex. include-directories)</param>
		/// <param name="shortForm">The short form of argument specifying this flag (ex. I)</param>
		/// <param name="hasValue">True if this flag requires an value, otherwise it is a simple boolean</param>
		public void Add(string key, string longForm, string shortForm, bool hasValue) {
			Add(key, longForm, shortForm, hasValue, string.Empty, string.Empty);
		}
		/// <summary>
		/// Adds new Flag into the collection
		/// </summary>
		/// <param name="key">Unique key of this flag</param>
		/// <param name="longForm">The expanded form of argument specifying this flag (ex. include-directories)</param>
		/// <param name="shortForm">The short form of argument specifying this flag (ex. I)</param>
		/// <param name="description">Description of this flag in automatically generated usage help</param>
		public void Add(string key, string longForm, string shortForm, string description) {
			Add(key, longForm, shortForm, false, string.Empty, description);
		}
		/// <summary>
		/// Returns Flag with specified key
		/// </summary>
		public Flag this[string key] {
			get {
				return (Flag)_innerHashtable[key];
			}
		}
		internal Flag GetByShortForm(string shortForm) {
			foreach(Flag flag in _innerHashtable.Values)
				if(shortForm==flag.ShortForm)
					return flag;

			return null;
		}
		internal Flag GetByLongForm(string longForm) {
			foreach(Flag flag in _innerHashtable.Values)
				if(longForm.StartsWith(flag.LongForm))
					return flag;

			return null;
		}
	}

	namespace Exceptions {
		internal class UnknownArgument : Exception {
			internal UnknownArgument(string argumentName) : base(string.Format("Unknown argument - {0}\r\n", argumentName)) {
			}
		}
		internal class MissingArgumentValue : Exception {
			internal MissingArgumentValue(string argumentName) : base(string.Format("Missing argument value - {0}", argumentName)) {
			}
		}
		public class HelpFlagReserved : Exception {
			internal HelpFlagReserved() : base("The flag --help is reserved for internal help generator") {
			}
		}
	}
}
