using System;
using System.Collections;
using System.Globalization;
using System.IO;
using System.Net;
using System.Text;
using System.Threading;
using System.Web;
using System.Xml;

namespace jachymko.Nyx {

	public class Witch : IDisposable {

		private const int MIN_REFRESH_INTERVAL = 10000;
		private const string WITCH_URI = "http://www.nyx.cz/code/client.php";
		private const string AUTH_STRING = "loguser={0}&logpass={1}";
		private const string SENDMAIL_STRING = AUTH_STRING + "&recipient={2}&message={3}";
		private static readonly Encoding ENCODING = Encoding.GetEncoding("ISO-8859-2");

		private string name = string.Empty;
		private string password = string.Empty;
		private string userAgent = "jachymko.Nyx.Witch";
		private int refreshInterval = 10000;

		private Thread workerThread = null;
		private AutoResetEvent authenticated = new AutoResetEvent(false);
		private XmlDocument document = null;
		private IWebProxy proxy = GlobalProxySelection.GetEmptyWebProxy();
		
		private Category[] prevCategories = null;
		private Friend[] previousActiveFriends = null;

		// WARNING!!!!
		// These events are not fired on the UI thread,
		// which means you have to do almost everything with
		// Windows.Forms using Control.BeginInvoke()
		public event FriendEventHandler FriendSignedIn;
		public event FriendEventHandler FriendSignedOut;
		public event NewMailEventHandler NewMail;
		public event NewWriteUpEventHandler NewWriteUp;
		public event EventHandler Disconnected;
		public event EventHandler Connected;
		

		public Witch(string name, string password) {
			if(name == null || password == null) {
				throw new ArgumentNullException();
			} else if(name.Length == 0 || password.Length == 0) {
				throw new ArgumentException();
			}

			this.name = name;
			this.password = password;
		}

		public Witch(string name, string password, WebProxy proxy) :
			this(name, password) {
			if(proxy != null) {
				this.proxy = proxy;
			}
		}

		public Witch(string name, string password, int refreshInterval) : this(name, password) {
			if(refreshInterval < MIN_REFRESH_INTERVAL) {
				throw new ArgumentOutOfRangeException("refreshInterval", refreshInterval,
					string.Format("refreshInterval must be greater than {0} ms", MIN_REFRESH_INTERVAL));
			}

			this.refreshInterval = refreshInterval;
		}

		public Witch(string name, string password, int refreshInterval, WebProxy proxy) : 
			this(name, password, refreshInterval) {
			if(proxy != null) {
				this.proxy = proxy;
			}
		}

		public void Dispose() {
			Close();
		}

		public void Open() {
			lock(this) {
				if(workerThread != null && workerThread.ThreadState != ThreadState.Aborted) {
					return;
				}

				authenticated.Reset();
				workerThread = new Thread(new ThreadStart(Worker));
				workerThread.Name = "Nyx.Witch.Worker()";
				workerThread.IsBackground = true;
				workerThread.Start();
			}

			// wait until first request completes
			if(authenticated.WaitOne(7500, false) == false) {
				throw new NetworkException("Timeout connecting nyx.cz");
			
			}
			
			ValidateObjectState();
		}

		public void Close() {
			document = null;

			lock(this) {
				if(workerThread != null && workerThread.ThreadState != ThreadState.Aborted) {
					lock(workerThread) {
						workerThread.Abort();
						workerThread = null;			
					}
				}
			}
		}

		public bool IsAuthenticated {
			get {
				lock(this) {
					if(document != null) {
						XmlNode node = document.SelectSingleNode("/info/page/error");

						if(node == null) {
							return true;
						}
					}
				}

				return  false;
			}
		}

		public CultureInfo Language {
			get {
				ValidateObjectState();
				XmlNode node = document.SelectSingleNode("/info/page/lang");

				if(node != null) {
					switch(node.InnerText.ToLower()) {
						case "cz":
							return new CultureInfo("cs");
					}
				}

				return new CultureInfo("en");
			}
		}

		public string Name {
			get {
				ValidateObjectState();
				XmlNode node = document.SelectSingleNode("/info/user/username");
				
				if(node != null) {
					return node.InnerText;
				}

				return string.Empty;
			}
		}

		public string UserAgent {
			get {
				return userAgent;
			}
			set {
				userAgent = value.Replace(' ', '_');
			}
		}

		public Friend[] ActiveFriends {
			get {
				if(document == null) return new Friend[0];

				ArrayList friends = new ArrayList();
				XmlNodeList nodes = document.SelectNodes("/info/friends/friend");

				if(nodes != null) {
					for(int i = 0; i < nodes.Count; i++) {
						friends.Add(new Friend(nodes[i]));
					}
				}

                return (Friend[])friends.ToArray(typeof(Friend));
			}
		}

		public Category[] Categories {
			get {
				ValidateObjectState();
				Category[] categories = new Category[0];

				XmlNodeList nodes = document.SelectNodes("/info/boards/category");
				if(nodes != null) {
					categories = new Category[nodes.Count];

					for(int i = 0; i < nodes.Count; i++) {
						categories[i] = new Category(nodes[i]);
					}
				}

				return categories;
			}
		}

		protected virtual void OnFriendSignedIn(FriendEventArgs args) {
			if(FriendSignedIn != null) {
				FriendSignedIn(this, args);
			}
		}

		protected virtual void OnFriendSignedOut(FriendEventArgs args) {
			if(FriendSignedOut != null) {
				FriendSignedOut(this, args);
			}
		}

		protected virtual void OnNewMail(NewMailEventArgs args) {
			if(NewMail != null) {
				NewMail(this, args);
			}
		}

		protected virtual void OnNewWriteUp(NewWriteUpEventArgs args) {
			if(NewWriteUp != null) {
				NewWriteUp(this, args);
			}
		}

		protected virtual void OnDisconnected(EventArgs args) {
			if(Disconnected != null) {
				Disconnected(this, args);
			}
		}
		protected virtual void OnConnected(EventArgs args) {
			if(Connected != null) {
				Connected(this, args);
			}
		}
		private void Worker() {
			while(true) {
				string content = string.Format(AUTH_STRING, HttpUtility.UrlEncode(name), HttpUtility.UrlEncode(password));

				HttpWebRequest request = (HttpWebRequest)HttpWebRequest.Create(WITCH_URI);
				request.Method = "POST";
				request.UserAgent = userAgent;
				request.Proxy = proxy;
		
				request.ContentLength = ENCODING.GetByteCount(content);
				request.ContentType = "application/x-www-form-urlencoded";
				
				try {
					using(StreamWriter sw = new StreamWriter(request.GetRequestStream(), ENCODING)) {
						sw.Write(content);
					}
				} catch {
					OnDisconnected(EventArgs.Empty);
					Thread.Sleep(refreshInterval);
					continue;
				}

				try {
					using(WebResponse response = request.GetResponse()) {
						using(Stream stream = response.GetResponseStream()) {
						
							XmlDocument doc = new XmlDocument();
							doc.Load(stream);

							lock(this) {
								this.document = doc;
							}

							// the Open() method waits on this AutoResetEvent
							// before returning
							authenticated.Set();

							if(!IsAuthenticated) {
								OnDisconnected(EventArgs.Empty);
								Thread.Sleep(refreshInterval);
								continue;
							}

							OnConnected(EventArgs.Empty);

							if(!IsCached) {
								UpdateMail();
							}

							FireEvents();
						}
					}
				} catch {
					document = null;
				}
				
				
				Thread.Sleep(refreshInterval);
			}
		}

		private void ValidateObjectState() {
			if(!IsAuthenticated) {
				//throw new WrongPasswordException();
			}
		}

		private void FireEvents() {
			if(previousActiveFriends != null) {
				Friend[] activeFriends = ActiveFriends;

				// look up friends which are offline now
				for(int i = 0; i < previousActiveFriends.Length; i++) {
					if(Array.IndexOf(activeFriends, previousActiveFriends[i]) < 0) {
						OnFriendSignedOut(new FriendEventArgs(previousActiveFriends[i]));
					}
				}

				// look up friends which were offline in the previous array
				for(int i = 0; i < activeFriends.Length; i++) {
					if(Array.IndexOf(previousActiveFriends, activeFriends[i]) < 0) {
						OnFriendSignedIn(new FriendEventArgs(activeFriends[i]));
					}
				}
			}
			previousActiveFriends = ActiveFriends;

			if(prevCategories != null) {
				ArrayList prevTopics = new ArrayList();
				ArrayList topics = new ArrayList();

				// fill the prevTopics array with all topics from previous request
				for(int i = 0; i < prevCategories.Length; i++) {
					prevTopics.AddRange((ICollection)prevCategories[i].Topics);
				}

				// fill the topics array with all current topics 
				Category[] cats = Categories;
				for(int i = 0; i < cats.Length; i++) {
					topics.AddRange((ICollection)cats[i].Topics);
				}

				// find topics missing in the prevTopics and topics which have topics[i].New > prevTopics[index].New;
				for(int i = 0; i < topics.Count; i++) {
					int index = prevTopics.IndexOf(topics[i]);
					
					if(index < 0) {
						OnNewWriteUp(new NewWriteUpEventArgs((Topic)topics[i]));
					} else if(index < prevTopics.Count) {
						Topic t = (Topic)prevTopics[index];

						if(((Topic)topics[i]).New > t.New) {
							OnNewWriteUp(new NewWriteUpEventArgs((Topic)topics[i]));
						}
					}
				}
			}
			prevCategories = Categories;
		}

		// cached response hasn't the mail block
		private bool IsCached {
			get {
				ValidateObjectState();
				XmlNode node = document.SelectSingleNode("/info/page/cached");

				if(node != null) {
					return node.InnerText.ToLower() == "yes" ? true : false;
				}

				return true;
			}
		}

		private void UpdateMail() {
			XmlNodeList nodes = document.SelectNodes("/info/mail/message");
			if(nodes != null) {
				for(int i = 0; i < nodes.Count; i++) {
					Mail msg = new Mail(nodes[i]);					
					OnNewMail(new NewMailEventArgs(msg));
				}
			}
		}
		public Mail SendMail(string recipient, string message) {
			Mail m = new Mail(name, message, DateTime.Now);

			string content = string.Format(SENDMAIL_STRING, HttpUtility.UrlEncode(name), HttpUtility.UrlEncode(password),
				HttpUtility.UrlEncode(recipient), HttpUtility.UrlEncode(message));

			HttpWebRequest request = (HttpWebRequest)HttpWebRequest.Create(WITCH_URI);
			request.Method = "POST";
			request.UserAgent = userAgent;
			request.Proxy = proxy;
		
			request.ContentLength = ENCODING.GetByteCount(content);
			request.ContentType = "application/x-www-form-urlencoded";
			
			try {
				using(StreamWriter sw = new StreamWriter(request.GetRequestStream(), ENCODING)) {
					sw.Write(content);
				}
			} catch {
				throw new NetworkException("could not connect");
			}


			XmlDocument doc = new XmlDocument();
			try {
				using(WebResponse response = request.GetResponse()) {
					using(Stream stream = response.GetResponseStream()) {
						doc.Load(stream);
					}
				}
			} catch {
				throw new NetworkException("could not connect");
			}
						
			XmlNode errorNode = doc.SelectSingleNode("/info/page/mailerror");
			if(errorNode != null) {
				throw new SendMailException(HttpUtility.HtmlDecode(errorNode.InnerText));
			}

			errorNode = document.SelectSingleNode("/info/page/error");
			if(errorNode != null) {
				throw new WrongPasswordException();
			}

			return m;
		}
	}

	public class WrongPasswordException : ApplicationException {

		internal WrongPasswordException() : base("Wrong username or password") {
		}
	}
	public class NetworkException : ApplicationException {
		internal NetworkException(string msg)  : base(msg) {
		}
	}
	public class SendMailException : ApplicationException {
		internal SendMailException(string msg) : base(msg) {
		}
	}

	public class FriendEventArgs : EventArgs {
		private Friend friend = Friend.Empty;
		
		public FriendEventArgs(Friend friend) {
			this.friend = friend;
		}

		public Friend Friend {
			get {
				return this.friend;
			}
		}
	}
	public class NewMailEventArgs : EventArgs {
		private Mail mail = null;

		public NewMailEventArgs(Mail mail) {
			this.mail = mail;
		}
		public Mail Mail {
			get {
				return this.mail;
			}
		}
	}
	public class NewWriteUpEventArgs : EventArgs {
		private Topic topic = null;

		public NewWriteUpEventArgs(Topic topic) {
			this.topic = topic;
		}
		public Topic Topic {
			get {
				return this.topic;
			}
		}
	}
	public delegate void FriendEventHandler(object sender, FriendEventArgs args);
	public delegate void NewMailEventHandler(object sender, NewMailEventArgs args);
	public delegate void NewWriteUpEventHandler(object sender, NewWriteUpEventArgs args);
}
