﻿using System;
using System.Drawing;
using System.Drawing.Drawing2D;
using System.Windows.Forms;

namespace jachymko.Nyx.Wokno {

	public class BookmarksListBox : ListBox {

		private Category[] bookmarks = new Category[0];
		private StringFormat catFormat = new StringFormat(StringFormatFlags.NoWrap);
		private StringFormat topicFormat = new StringFormat(StringFormatFlags.NoWrap);
		private int unread = 0;

		public BookmarksListBox() {

			DrawMode = DrawMode.OwnerDrawVariable;
			IntegralHeight = false;

			BackColor = Color.FromArgb(0xb8, 0xb9, 0x92);

			SetStyle(ControlStyles.AllPaintingInWmPaint, true);
			SetStyle(ControlStyles.Opaque, true);
			SetStyle(ControlStyles.ResizeRedraw, true);

			catFormat.Alignment = StringAlignment.Center;
			topicFormat.LineAlignment = catFormat.LineAlignment = StringAlignment.Center;
			topicFormat.Trimming = catFormat.Trimming = StringTrimming.EllipsisCharacter;

		}

		public Category[] Bookmarks {
			get {
				return bookmarks;
			}
			set {
				bookmarks = value;
				
				BeginUpdate();
				Items.Clear();
				unread = 0;

				for(int i = 0; i < bookmarks.Length; i++) {
					Items.Add(bookmarks[i]);

					for(int j = 0; j < bookmarks[i].Topics.Length; j++) {
						Items.Add(bookmarks[i].Topics[j]);
						unread += bookmarks[i].Topics[j].New;
					}
				}
				EndUpdate();
			}
		}
		public int Unread {
			get {
				return unread;
			}
		}

		private void WatchItem(int index) {

		}

		protected override void OnMeasureItem(MeasureItemEventArgs e) {
			e.ItemWidth = ClientSize.Width;

			if(Items[e.Index] is Category) {
				Category cat = (Category)Items[e.Index];

				using(Font fnt = new Font(Font, FontStyle.Bold)) {
					e.ItemHeight = 4 + (int)e.Graphics.MeasureString(cat.Name, fnt, ClientSize.Width, catFormat).Height;
				}
			} else if(Items[e.Index] is Topic) {
				Topic t = (Topic)Items[e.Index];

				e.ItemHeight = 4 + (int)e.Graphics.MeasureString(t.Name, Font, ClientSize.Width).Height;
			}
		}


		protected override void OnMouseDown(MouseEventArgs e) {
			int idx = -1;

			for(int i = 0; i < Items.Count; i++) {
				Rectangle rect = GetItemRectangle(i);

				if(ClientRectangle.IntersectsWith(rect) && rect.Contains(e.X, e.Y)) {
					idx = i;
					break;
				}
			}

			SelectedIndex = idx;
			base.OnMouseDown(e);
		}

		protected override void OnDrawItem(DrawItemEventArgs e) {
			if(e.Index < 0 || Items == null || Items.Count == 0 ||
				e.Bounds.Width == 0 || e.Bounds.Height == 0) return;

			using(Bitmap offscreen = new Bitmap(e.Bounds.Width, e.Bounds.Height, e.Graphics)) {
				using(Graphics g = Graphics.FromImage(offscreen)) {
					g.SmoothingMode = SmoothingMode.HighQuality;
					g.TextRenderingHint = System.Drawing.Text.TextRenderingHint.ClearTypeGridFit;

					Rectangle rect = new Rectangle(0, 0, e.Bounds.Width, e.Bounds.Height);
					bool selected = ((e.State & DrawItemState.Selected) == DrawItemState.Selected) && Focused;


					if(Items[e.Index] is Category) {
						Category c = (Category)Items[e.Index];

						if(selected) {
							g.Clear(SystemColors.Highlight);
						} else {
							g.Clear(Color.FromArgb(0x87, 0x93, 0x7b));
						}

//						rect.Y += 4;
//						rect.Height -= 4;

						string name = ":: " + System.Web.HttpUtility.HtmlDecode(c.Name) + " ::";

						using(Font fnt = new Font(Font, FontStyle.Bold)) {
							Brush b = (selected ? SystemBrushes.HighlightText : Brushes.White);
							g.DrawString(name, fnt, b, rect, catFormat);
						}
					} else if(Items[e.Index] is Topic) {
						Topic t = (Topic)Items[e.Index];

						if(selected) {
							g.Clear(SystemColors.Highlight);
						} else {
							g.Clear(Color.FromArgb(0xb8, 0xb9, 0x92));
						}

						string n = "NEW: " + ((t.New >= 100) ? "100+" : t.New.ToString());
						SizeF newSize = g.MeasureString(n, Font, ClientSize.Width / 2);

						rect.Width -= (int)newSize.Width;

						using(Font fnt = new Font(Font, (Startup.WatchTopic[t.Id] ? FontStyle.Bold : FontStyle.Regular))) {
							string name = System.Web.HttpUtility.HtmlDecode(t.Name);
							SizeF nameSize = g.MeasureString(name, fnt, rect.Width, topicFormat);

							Brush b = (selected ? SystemBrushes.HighlightText : SystemBrushes.WindowText);
						
							g.DrawString(name, fnt, b, rect, topicFormat);
						

							rect.X = (int)nameSize.Width;
							rect.Width = (int)newSize.Width + 5;

							b = (selected ? SystemBrushes.HighlightText : Brushes.Red);
							g.DrawString(n, Font, b, rect, topicFormat);
						}
					}

					e.Graphics.DrawImageUnscaled(offscreen, e.Bounds.Location);
				}
			}
		}
	}
}
