using System;
using System.Collections.Specialized;
using System.Drawing;
using System.Drawing.Imaging;
using System.Windows.Forms;

namespace jachymko.Nyx.Wokno {

	public class BufferedListBox : Panel {
	
		private string[] items = new string[] { "jedna", "dva", "tri", "ctyri", "pet",
												  "cos", "to", "janku", "cos", "to", "sned",
												  "bezi", "liska", "k", "taboru", "nese", 
												  "pytel", "zazvoru", "jezek", "za", "ni",
												  "pospicha", "ze", "ji", "pytel", "rozpicha"
											  };
		private Graphics graphics = null;
		private Bitmap offscreen = null;
		private Rectangle[] itemRects = new Rectangle[0];
		private int selectedIndex = -1;
		private VScrollBar scrollBar = new VScrollBar();

		public BufferedListBox() {
			SetStyle(ControlStyles.UserPaint, true);
			SetStyle(ControlStyles.AllPaintingInWmPaint, true);
			SetStyle(ControlStyles.Selectable, true);
			SetStyle(ControlStyles.Opaque, true);
			SetStyle(ControlStyles.ResizeRedraw, true);

			BackColor = SystemColors.Window;
			ForeColor = SystemColors.WindowText;

			scrollBar.Location = new Point(ClientSize.Width -
				SystemInformation.VerticalScrollBarWidth, 0);
			scrollBar.Size = new Size(SystemInformation.VerticalScrollBarWidth,
				ClientSize.Height);
			scrollBar.Scroll += new ScrollEventHandler(OnScroll);
			scrollBar.Dock = DockStyle.Right;

			scrollBar.LargeChange = 20;

			Controls.Add(scrollBar);
		}

		public int ItemHeight {
			get {
				return 20;
			}
			set {
				//TODO: guess ;)
			}
		}

		public int SelectedIndex {
			get {
				return selectedIndex;
			}
			set {
				SelectItem(value);
			}
		}

		public int GetItemAtPoint(Point pt) {
			if(scrollBar.Visible) {
				pt.Y += scrollBar.Value;
			}

			for(int i = 0; i < itemRects.Length; i++) {
				if(itemRects[i].Contains(pt)) {
					return i;
				}
			}

			return -1;
		}

		protected override void OnHandleCreated(EventArgs e) {
			base.OnHandleCreated (e);

			Recalculate();
		}

		protected override void OnResize(EventArgs e) {
			Recalculate();

			base.OnResize (e);
		}

		protected override void OnMouseDown(MouseEventArgs e) {
			SelectItem(GetItemAtPoint(new Point(e.X, e.Y)));
			base.OnMouseDown (e);
		}

		/*protected override void OnMouseUp(MouseEventArgs e) {
			SelectItem(GetItemAtPoint(new Point(e.X, e.Y)));

			base.OnMouseDown (e);
		}*/

		protected override void OnMouseMove(MouseEventArgs e) {
			if(e.Button == MouseButtons.Left) {
				SelectItem(GetItemAtPoint(new Point(e.X, e.Y)));
			}
			base.OnMouseMove (e);
		}

		protected override bool IsInputKey(Keys keyData) {
			if(keyData == Keys.Up || keyData == Keys.Down) {
				return false;
			} else {
				return base.IsInputKey (keyData);
			}
		}

		protected override bool ProcessDialogKey(Keys keyData) {
			if(keyData == Keys.Up && selectedIndex > 0) {
				SelectItem(selectedIndex - 1);
				return true;
			}
			if(keyData == Keys.Down && selectedIndex < (items.Length - 1)) {
				SelectItem(selectedIndex + 1);
				return true;
			}

			return base.ProcessDialogKey (keyData);
		}

		protected override void OnPaint(PaintEventArgs e) {
			base.OnPaint (e);

			graphics.Clear(BackColor);
			for(int i = 0; i < items.Length; i++) {
				DrawItemEventArgs args = new DrawItemEventArgs(graphics, Font, itemRects[i], i, 
					(i == selectedIndex ? DrawItemState.Selected : DrawItemState.Default), ForeColor, BackColor);
				OnDrawItem(args);
			}

			if(!scrollBar.Visible) {
				e.Graphics.DrawImage(offscreen, 0, 0);
			} else {
				e.Graphics.DrawImage(offscreen, 0, -scrollBar.Value);
			}
		}

		protected virtual void OnDrawItem(DrawItemEventArgs e) {
			bool selected = ((e.State & DrawItemState.Selected) == DrawItemState.Selected);

			Brush backBrush = selected ? SystemBrushes.Highlight : SystemBrushes.Window;
			Brush foreBrush = selected ? SystemBrushes.HighlightText : SystemBrushes.WindowText;

			e.Graphics.FillRectangle(backBrush, e.Bounds);
			e.Graphics.DrawString(items[e.Index], Font, foreBrush, e.Bounds);
		}

		protected virtual void OnMeasureItem(MeasureItemEventArgs e) {
			e.ItemWidth = ClientSize.Width - ScrollBarWidth;
		}

		private void SelectItem(int index) {
			if(index < -1 || index >= items.Length) {
				throw new ArgumentOutOfRangeException("index");
			}

			selectedIndex = index;
			ScrollToSelected();
			Invalidate();
		}

		private void ScrollToSelected() {
			if(selectedIndex < 0) return;

			Rectangle rect = itemRects[selectedIndex];
			rect.Offset(AutoScrollPosition);


			if(rect.Top < 0) {
				AutoScrollPosition.Offset(0, -rect.Top);
			}


		}

		private void Recalculate() {
			if(graphics != null) {
				graphics.Dispose();
			}

			if(offscreen != null) {
				offscreen.Dispose();
			}

			Size totalSize = CalculateLayout();
			
			scrollBar.Visible = (totalSize.Height > ClientSize.Height);
//			if(!scrollBar.Visible)

			offscreen = new Bitmap(totalSize.Width, totalSize.Height, PixelFormat.Format32bppArgb);
			graphics = Graphics.FromImage(offscreen);
		}

		private Size CalculateLayout() {
			Size totalSize = Size.Empty;
			itemRects = new Rectangle[items.Length];
			int y = 0;

			for(int i = 0; i < items.Length; i++) {
				MeasureItemEventArgs e = new MeasureItemEventArgs(graphics, i, ItemHeight);
				OnMeasureItem(e);

				totalSize.Width = Math.Max(e.ItemWidth, totalSize.Width);
				totalSize.Width = (totalSize.Width > (ClientSize.Width  - ScrollBarWidth)) ? 
					(ClientSize.Width  - ScrollBarWidth) : totalSize.Width ; 
				totalSize.Height += e.ItemHeight;

				itemRects[i].X = 0;
				itemRects[i].Y = y;
				itemRects[i].Width = totalSize.Width;
				itemRects[i].Height = e.ItemHeight;

				y += e.ItemHeight;
			}

			scrollBar.Minimum = 0;
			scrollBar.Maximum = totalSize.Height - ClientSize.Height;

			return totalSize;
		}

		private void OnScroll(object sender, ScrollEventArgs e) {
			Invalidate();
		}

		private int ScrollBarWidth {
			get {
				return scrollBar.Visible ? scrollBar.Width : 0;
			}
		}
	}
}
