using System;
using System.Collections;
using System.ComponentModel;
using System.Diagnostics;
using System.Drawing;
using System.IO;
using System.Net;
using System.Text;
using System.Web;
using System.Xml;

using Microsoft.Win32;

namespace jachymko.Nyx.Wokno {

	public class Config {

		public enum NotificationType {
			ShowWindow,
			Balloon,
			Icon,
			None
		}

		private string fileName = string.Empty;
		private XmlDocument doc = null;

		private bool encryptMailCache = DataProtection.IsSupported;
		private bool encryptHistory = DataProtection.IsSupported;
		private bool encryptPasswords = DataProtection.IsSupported;
		private string userName = string.Empty;
		private string password = string.Empty;
		private string webPassword = string.Empty;
		private int refreshInterval = 40000;
		private FriendsListBoxView iconSize;
		private bool showAltTime = true;
		private bool alwaysOnTop = false;
		private int opacity = 100;
		private bool showInTaskbar = false;
		private Point location = Point.Empty;
		private Size size = Size.Empty;
		private NotificationType mailNotify = NotificationType.Icon;
		private NotificationType friendOnlineNotify = NotificationType.Balloon;
		private NotificationType friendOfflineNotify = NotificationType.None;
		private NotificationType writeupNotify = NotificationType.Balloon;
		private AllFriends allFriends = new AllFriends();
		private WebProxy proxy = WebProxy.GetDefaultProxy();
		private Point historyLocation, bookLocation;
		private Size historySize, bookSize;
		private bool unknownOnline = true;
		private bool unknownOffline = true;
		private bool autoResize = false;
		private double autoResizeMax = 0.75;
		private bool autoUpdate = true;
		private ArrayList userModes = new ArrayList();
		private bool autoAwayOnScreensaver = true;
		private bool autoAwayOnLock = true;
		private int autoAwayDelay = (5 * 60); // 5 minutes
		private string lastMode = string.Empty;
		private int autoUpdateInterval = 2;

		public Config(string fileName) {
			this.fileName = fileName;

			if(File.Exists(fileName)) {
				Read();
			}
		}

		public void Read() {
			doc = new XmlDocument();

			XmlTextReader xtr = new XmlTextReader(File.Open(fileName, FileMode.Open, FileAccess.Read, FileShare.Read));
			try {
				doc.Load(xtr);
			} catch {
				xtr.Close();
				return;
			}
			xtr.Close();

			XmlNode n = doc.SelectSingleNode("/config/nyx");
			if(n != null) {
				if(n.Attributes["encrypted"] != null) encryptPasswords = bool.Parse(n.Attributes["encrypted"].InnerText);
				if(n.Attributes["username"] != null) userName = n.Attributes["username"].InnerText;
				if(n.Attributes["password"] != null) password = DecryptBase64(n.Attributes["password"].InnerText);
				if(n.Attributes["webpassword"] != null) webPassword = DecryptBase64(n.Attributes["webpassword"].InnerText);
				if(n.Attributes["refresh"] != null) RefreshInterval = int.Parse(n.Attributes["refresh"].InnerText);
			}

			n = doc.SelectSingleNode("/config/proxy");
			if(n != null) {
				string address = string.Empty;
				int port = 0;

				if(n.Attributes["address"] != null) address = n.Attributes["address"].InnerText;
				if(n.Attributes["port"] != null) port = int.Parse(n.Attributes["port"].InnerText);
				
				if(address.Length > 0 && port != 0) {
					proxy = new WebProxy(address, port);
				}
			}

			n = doc.SelectSingleNode("/config/usermodes");
			if(n != null) {
				if(n.Attributes["awayOnSaver"] != null) autoAwayOnScreensaver = bool.Parse(n.Attributes["awayOnSaver"].InnerText);
				if(n.Attributes["awayOnLock"] != null) autoAwayOnLock = bool.Parse(n.Attributes["awayOnLock"].InnerText);
				if(n.Attributes["awayDelay"] != null) autoAwayDelay = int.Parse(n.Attributes["awayDelay"].InnerText);
				if(n.Attributes["lastMode"] != null) lastMode = n.Attributes["lastMode"].InnerText;

				XmlNodeList l = n.ChildNodes;
				if(l != null && l.Count > 0) {
					for(int i = 0; i < l.Count; i++) {
						if(l[i] != null) {
							userModes.Add(l[i].InnerText);
						}
					}
				}
			}

			n = doc.SelectSingleNode("/config/window");
			if(n != null) {
				if(n.Attributes["autoupdate"] != null) autoUpdate = bool.Parse(n.Attributes["autoupdate"].InnerText);
				if(n.Attributes["auinterval"] != null) autoUpdateInterval = int.Parse(n.Attributes["auinterval"].InnerText);
				if(n.Attributes["iconsize"] != null) iconSize = (FriendsListBoxView)Enum.Parse(typeof(FriendsListBoxView), n.Attributes["iconsize"].InnerText);
				if(n.Attributes["alttime"] != null) showAltTime = bool.Parse(n.Attributes["alttime"].InnerText);
				if(n.Attributes["ontop"] != null) alwaysOnTop = bool.Parse(n.Attributes["ontop"].InnerText);
				if(n.Attributes["opacity"] != null) opacity = int.Parse(n.Attributes["opacity"].InnerText);
				if(n.Attributes["taskbar"] != null) showInTaskbar = bool.Parse(n.Attributes["taskbar"].InnerText);
				if(n.Attributes["location"] != null) location = (Point)TypeDescriptor.GetConverter(typeof(Point)).ConvertFromInvariantString(n.Attributes["location"].InnerText);
				if(n.Attributes["size"] != null) size = (Size)TypeDescriptor.GetConverter(typeof(Size)).ConvertFromInvariantString(n.Attributes["size"].InnerText);
				if(n.Attributes["autoresize"] != null) autoResize = bool.Parse(n.Attributes["autoresize"].InnerText);
				if(n.Attributes["maxautosize"] != null) autoResizeMax = double.Parse(n.Attributes["maxautosize"].InnerText);
			}

			n = doc.SelectSingleNode("/config/window/history");
			if(n != null) {
				if(n.Attributes["location"] != null) historyLocation = (Point)TypeDescriptor.GetConverter(typeof(Point)).ConvertFromInvariantString(n.Attributes["location"].InnerText);
				if(n.Attributes["size"] != null) historySize = (Size)TypeDescriptor.GetConverter(typeof(Size)).ConvertFromInvariantString(n.Attributes["size"].InnerText);
			}

			n = doc.SelectSingleNode("/config/window/bookmarks");
			if(n != null) {
				if(n.Attributes["location"] != null) bookLocation = (Point)TypeDescriptor.GetConverter(typeof(Point)).ConvertFromInvariantString(n.Attributes["location"].InnerText);
				if(n.Attributes["size"] != null) bookSize = (Size)TypeDescriptor.GetConverter(typeof(Size)).ConvertFromInvariantString(n.Attributes["size"].InnerText);
			}

			n = doc.SelectSingleNode("/config/notifications");
			if(n != null) {
				if(n.Attributes["mail"] != null) mailNotify = (NotificationType)Enum.Parse(typeof(NotificationType), n.Attributes["mail"].InnerText, true);
				if(n.Attributes["friendOnline"] != null) friendOnlineNotify = (NotificationType)Enum.Parse(typeof(NotificationType), n.Attributes["friendOnline"].InnerText, true);
				if(n.Attributes["friendOffline"] != null) friendOfflineNotify = (NotificationType)Enum.Parse(typeof(NotificationType), n.Attributes["friendOffline"].InnerText, true);
				if(n.Attributes["unknownFriendOnline"] != null) unknownOnline = bool.Parse(n.Attributes["unknownFriendOnline"].InnerText);
				if(n.Attributes["unknownFriendOffline"] != null) unknownOffline = bool.Parse(n.Attributes["unknownFriendOffline"].InnerText);
			}

			n = doc.SelectSingleNode("/config/friends");
			if(n != null) {
				allFriends = new AllFriends(n);
			}

			n = doc.SelectSingleNode("/config/watch");
			if(n != null) {
				Startup.WatchTopic.Load(n);
			}

		}

		public void Save() {
			XmlTextWriter xtw = new XmlTextWriter(fileName, Encoding.UTF8);
			xtw.Formatting = Formatting.Indented;

			xtw.WriteStartDocument(true);
			xtw.WriteStartElement("config");	// /config
			
			xtw.WriteStartElement("nyx");		// /config/nyx
			xtw.WriteAttributeString("encrypted", encryptPasswords.ToString());
			xtw.WriteAttributeString("username", userName);
			xtw.WriteAttributeString("password", EncryptBase64(password));
			xtw.WriteAttributeString("webpassword", EncryptBase64(webPassword));
			xtw.WriteAttributeString("refresh", refreshInterval.ToString());
			xtw.WriteEndElement();				// /config/nyx	

			if(proxy != null && proxy.Address != null) {
				xtw.WriteStartElement("proxy");	// /config/proxy
				xtw.WriteAttributeString("address", proxy.Address.Host);
				xtw.WriteAttributeString("port", proxy.Address.Port.ToString());
				xtw.WriteEndElement();			// /config/proxy
			}

			xtw.WriteStartElement("usermodes");

			xtw.WriteAttributeString("awayOnSaver", autoAwayOnScreensaver.ToString());
			xtw.WriteAttributeString("awayOnLock", autoAwayOnLock.ToString());
			xtw.WriteAttributeString("awayDelay", autoAwayDelay.ToString());
			xtw.WriteAttributeString("lastMode", lastMode);

			for(int i = 0; i < userModes.Count; i++) {
				xtw.WriteElementString("mode", (string)userModes[i]);
			}
			xtw.WriteEndElement();

			xtw.WriteStartElement("window");	// /config/window
			xtw.WriteAttributeString("autoupdate", autoUpdate.ToString());
			xtw.WriteAttributeString("auinterval", autoUpdateInterval.ToString());
			xtw.WriteAttributeString("iconsize", iconSize.ToString());
			xtw.WriteAttributeString("alttime", showAltTime.ToString());
			xtw.WriteAttributeString("ontop", alwaysOnTop.ToString());
			xtw.WriteAttributeString("opacity", opacity.ToString());
			xtw.WriteAttributeString("taskbar", showInTaskbar.ToString());
			xtw.WriteAttributeString("location", TypeDescriptor.GetConverter(location).ConvertToInvariantString(location));
			xtw.WriteAttributeString("size", TypeDescriptor.GetConverter(size).ConvertToInvariantString(size));
			xtw.WriteAttributeString("autoresize", autoResize.ToString());
			xtw.WriteAttributeString("maxautosize", autoResizeMax.ToString());

			xtw.WriteStartElement("history");	// /config/window/history
			xtw.WriteAttributeString("location", TypeDescriptor.GetConverter(historyLocation).ConvertToInvariantString(historyLocation));
			xtw.WriteAttributeString("size", TypeDescriptor.GetConverter(historySize).ConvertToInvariantString(historySize));
			xtw.WriteEndElement();				// /config/window/history

			xtw.WriteStartElement("bookmarks");	// /config/window/bookmarks
			xtw.WriteAttributeString("location", TypeDescriptor.GetConverter(bookLocation).ConvertToInvariantString(bookLocation));
			xtw.WriteAttributeString("size", TypeDescriptor.GetConverter(bookSize).ConvertToInvariantString(bookSize));
			xtw.WriteEndElement();				// /config/window/bookmarks
			xtw.WriteEndElement();				// /config/window


			xtw.WriteStartElement("notifications"); // /config/notifications
			xtw.WriteAttributeString("mail", mailNotify.ToString());
			xtw.WriteAttributeString("friendOnline", FriendOnlineNotify.ToString());
			xtw.WriteAttributeString("friendOffline", friendOfflineNotify.ToString());
			xtw.WriteAttributeString("unknownFriendOnline", unknownOnline.ToString());
			xtw.WriteAttributeString("unknownFriendOffline", unknownOffline.ToString());
			xtw.WriteEndElement();					// /config/notifications

			allFriends.Save(xtw);
			Startup.WatchTopic.Save(xtw);

			xtw.WriteEndElement();				// /config
			xtw.WriteEndDocument();
			xtw.Close();
		}

		private string DecryptBase64(string base64) {
			byte[] cipher = Convert.FromBase64String(base64);
			byte[] utf8 = new byte[0];

			if(encryptPasswords) {
				DataProtection dp = new DataProtection(DataProtection.Store.USE_USER_STORE);
				utf8 = dp.Decrypt(cipher);
			} else {
				utf8 = cipher;
			}

			return Encoding.UTF8.GetString(utf8);
		}
		private string EncryptBase64(string plain) {
			byte[] utf8 = Encoding.UTF8.GetBytes(plain);
			byte[] cipher = new byte[0];

			if(encryptPasswords) {
				DataProtection dp = new DataProtection(DataProtection.Store.USE_USER_STORE);
				cipher = dp.Encrypt(utf8);
			} else {
				cipher = utf8;
			}

			return Convert.ToBase64String(cipher);
		}

		public bool EncryptPassword {
			get {
				return encryptPasswords;
			}
			set {
				encryptPasswords = value;
			}
		}
		public bool ShowAltTime {
			get {
				return showAltTime;
			}
			set {
				showAltTime = value;
			}
		}
		public bool EncryptMailCache {
			get {
				return encryptMailCache;
			}
			set {
				encryptMailCache = value;
			}
		}
		public int RefreshInterval {
			get {
				return refreshInterval;
			}
			set {
				if(value < 10000) throw new ArgumentOutOfRangeException("value", value, "refreshInterval must be greater than 10000");

				refreshInterval = value;
			}
		}
		public string UserName {
			get {
				return this.userName;
			}
			set {
				this.userName = value;
			}
		}
		public string Password {
			get {
				return this.password;
			}
			set {
				this.password = value;
			}
		}

		public FriendsListBoxView IconSize {
			get {
				return iconSize;
			}
			set {
				iconSize = value;
			}
		}
		public bool AlwaysOnTop {
			get {
				return alwaysOnTop;
			}
			set {
				alwaysOnTop = value;
			}
		}
		public int Opacity {
			get {
				return opacity;
			}
			set {
				opacity = value;
			}
		}
		public NotificationType MailNotify {
			get {
				return mailNotify;
			}
			set {
				mailNotify = value;
			}
		}
		public NotificationType FriendOnlineNotify {
			get {
				return friendOnlineNotify;
			}
			set {
				if(value == NotificationType.ShowWindow) {
					throw new ArgumentException("FriendOnlineNotify cannot be NotificationType.ShowWindow");
				}

				friendOnlineNotify = value;
			}
		}
		public NotificationType FriendOfflineNotify {
			get {
				return friendOfflineNotify;
			}
			set {
				if(value == NotificationType.ShowWindow) {
					throw new ArgumentException("FriendOfflineNotify cannot be NotificationType.ShowWindow");
				}

				friendOfflineNotify = value;
			}
		}
		public NotificationType WriteupNotify {
			get {
				return writeupNotify;
			}
			set {
				if(value == NotificationType.ShowWindow) {
					throw new ArgumentException("WriteupNotify cannot be NotificationType.ShowWindow");
				}

				writeupNotify = value;
			}
		}
		public Point Location {
			get {
				return location;
			}
			set {
				location = value;
			}
		}
		public Point HistoryLocation {
			get {
				return historyLocation;
			}
			set {
				historyLocation = value;
			}
		}
		public Point BookLocation {
			get {
				return bookLocation;
			}
			set {
				bookLocation = value;
			}
		}
		public Size HistorySize {
			get {
				return historySize;
			}
			set {
				historySize = value;
			}
		}
		public Size BookSize {
			get {
				return bookSize;
			}
			set {
				bookSize = value;
			}
		}
		public Size Size {
			get {
				return size;
			}
			set {
				size = value;
			}
		}
		public bool ShowInTaskbar {
			get {
				return showInTaskbar;
			}
			set {
				showInTaskbar = value;
			}
		}

		public AllFriends AllFriends {
			get {
				return allFriends;
			}
		}
		public WebProxy Proxy {
			get {
				return proxy;
			}
			set {
				proxy = value;
			}
		}
		public bool EncryptHistory {
			get {
				return encryptHistory;
			}
			set {
				encryptHistory = value;
			}
		}

		public bool RunAtStartup {
			get {
				using(RegistryKey runKey = Registry.CurrentUser.CreateSubKey("SOFTWARE\\Microsoft\\Windows\\CurrentVersion\\Run")) {
					string val = ((string)runKey.GetValue("NyxWokno", string.Empty)).ToLower();

					return val == Process.GetCurrentProcess().MainModule.FileName.ToLower();
				}
			}
			set {
				using(RegistryKey runKey = Registry.CurrentUser.CreateSubKey("SOFTWARE\\Microsoft\\Windows\\CurrentVersion\\Run")) {
					if(value) {
						runKey.SetValue("NyxWokno", Process.GetCurrentProcess().MainModule.FileName);
					} else {
						runKey.DeleteValue("NyxWokno", false);
					}
				}
			}
		}
		public string WebPassword {
			get {
				return webPassword;
			}
			set {
				webPassword = value;
			}
		}
		public bool NotifyUnknownOffline {
			get {
				return unknownOffline;
			}
			set {
				unknownOffline = value;
			}
		}
		public bool NotifyUnknownOnline {
			get {
				return unknownOnline;
			}
			set {
				unknownOnline = value;
			}
		}

		public bool AutoResize {
			get {
				return autoResize;
			}
			set {
				autoResize = value;
			}
		}
		public double AutoResizeMax {
			get {
				return autoResizeMax;
			}
			set {
				autoResizeMax = value;
			}
		}
		public bool AutoUpdate {
			get {
				return this.autoUpdate;
			}
			set {
				this.autoUpdate = value;
			}
		}
		public int AutoUpdateInterval {
			get {
				return autoUpdateInterval;
			}
			set {
				autoUpdateInterval = value;
			}
		}
		public ArrayList UserModes {
			get {
				return userModes;
			}
		}

		public bool AutoAwayOnScreensaver {
			get {
				return autoAwayOnScreensaver;
			}
			set {
				autoAwayOnScreensaver = value;
			}
		}

		public bool AutoAwayOnLock {
			get {
				return autoAwayOnLock;
			}
			set {
				autoAwayOnLock = value;
			}
		}
		
		public int AutoAwayDelay {
			get {
				return autoAwayDelay;
			}
			set {
				autoAwayDelay = value;
			}
		}

		public string LastStatus {
			get {
				return lastMode;
			}
			set {
				lastMode = value;
			}
		}
	}

}
