using System;
using System.Collections;
using System.Diagnostics;
using System.Drawing;
using System.Drawing.Drawing2D;
using System.Drawing.Imaging;
using System.IO;
using System.Net;
using System.Reflection;
using System.Windows.Forms;

namespace jachymko.Nyx.Wokno {

	public enum FriendsListBoxView {
		LargeIcons,
		SmallIcons,
		NoIcons
	}

	public class FriendsListBox : ListBox {

		private readonly StringFormat itemFormat;
		internal static readonly Image mailIcon;
		

		private FriendsListBoxView view;
		private Image[] icons = new Image[0];
		private Friend[] friends = new Friend[0];
		private bool[] hasMail = new bool[0];

		static FriendsListBox() {
			using(Stream stream = Assembly.GetExecutingAssembly().GetManifestResourceStream("jachymko.Nyx.Wokno.Mail.gif")) {
				mailIcon = new Bitmap(stream);
			}
		}

		public FriendsListBox() {
			itemFormat =  new StringFormat(StringFormatFlags.NoWrap);
			itemFormat.LineAlignment = StringAlignment.Center;

			DrawMode = DrawMode.OwnerDrawFixed;
			IntegralHeight = false;

			SetStyle(ControlStyles.AllPaintingInWmPaint, true);
			SetStyle(ControlStyles.Opaque, true);
			SetStyle(ControlStyles.ResizeRedraw, true);
		}

		public Friend[] Friends {
			get {
				return this.friends;
			}
			set {
				BeginUpdate();

				int oldSelIdx = SelectedIndex;

				if(value == null || value.Length == 0) {				
					Items.Clear();
					this.friends = value;
				} else {
					View = view; // reset item height
					this.friends = value;
					LoadIcons();
				
					Items.Clear();
					for(int i = 0; i < friends.Length; i++) {
						Items.Add( friends[i] );
					}
				}

				if(oldSelIdx < Items.Count) {
					SelectedIndex = oldSelIdx;
				}

				EndUpdate();

			}
		}

		public FriendsListBoxView View {
			get {
				return this.view;
			}
			set {
				this.view = value;

				int minHeight = int.MinValue;
				using(Graphics g = CreateGraphics()) {
					minHeight = 4 + g.MeasureString("BLAH-BLAH", Font, int.MaxValue, itemFormat).ToSize().Height;
				}

				switch(this.view) {
					case FriendsListBoxView.NoIcons:
						ItemHeight = minHeight;
						break;
					case FriendsListBoxView.LargeIcons:
						ItemHeight = Math.Max(50, minHeight);
						break;
					default:
						ItemHeight = Math.Max(25, minHeight);
						break;
				}
			}
		}

		public void SetMail(string userName, bool value) {
			for(int i = 0; i < friends.Length; i++) {
				if(friends[i].UserName == userName.ToUpper()) {
					hasMail[i] = value;
					return;
				}
			}
		}
//		
//		protected override void OnPaintBackground(PaintEventArgs pevent) {
//			return;
//		}
//
		protected override void OnDrawItem(DrawItemEventArgs e) {
			if(e.Index < 0 || friends == null || friends.Length == 0 ||
				e.Bounds.Width == 0 || e.Bounds.Height == 0) return;

			using(Bitmap offscreen = new Bitmap(e.Bounds.Width, e.Bounds.Height, e.Graphics)) {
				using(Graphics g = Graphics.FromImage(offscreen)) {
					g.SmoothingMode = SmoothingMode.HighQuality;
					g.TextRenderingHint = System.Drawing.Text.TextRenderingHint.ClearTypeGridFit;

					Rectangle rect = new Rectangle(0, 0, e.Bounds.Width, e.Bounds.Height);
					bool selected = ((e.State & DrawItemState.Selected) == DrawItemState.Selected) && (Form.FromChildHandle(Handle).Focused);

					if(selected) {
						g.Clear(SystemColors.Highlight);
					} else {
						g.Clear(SystemColors.Window);
					}

					switch(view) {
						case FriendsListBoxView.NoIcons:
							rect = new Rectangle(new Point(0, 0), e.Bounds.Size);
							DrawUserName(e.Index, rect, g, selected);
							break;
						case FriendsListBoxView.LargeIcons:
							rect = new Rectangle(new Point(0, 0), new Size(40, 50));
							DrawIcon(e.Index, rect, g, selected);
				
							rect = new Rectangle(new Point(40, 0), e.Bounds.Size);
							rect.Height = rect.Height / 2;
							DrawUserName(e.Index, rect, g, selected);

							rect.Y += rect.Height;
							DrawLastRefresh(e.Index, rect, g, selected);
							break;
						default: // FriendsListBoxView.SmallIcons
							rect = new Rectangle(new Point(0, 0), new Size(20, 25));
							DrawIcon(e.Index, rect, g, selected);

							rect = new Rectangle(new Point(20, 0), e.Bounds.Size);
							DrawUserName(e.Index, rect, g, selected);
							break;

					}

					e.Graphics.DrawImageUnscaled(offscreen, e.Bounds.Location);
				}
			}
		}

		private void DrawUserName(int index, Rectangle bounds, Graphics g, bool selected) {
			Brush b = (selected ? SystemBrushes.HighlightText : SystemBrushes.WindowText);
			g.DrawString(friends[index].UserName, Font, b, bounds, itemFormat);
		}

		private void DrawIcon(int index, Rectangle bounds, Graphics g, bool selected) {
			g.DrawImage(icons[index], bounds);

			if(selected) {
				using(Brush b = new SolidBrush(Color.FromArgb(100, SystemColors.Highlight))) {
					g.FillRectangle(b, bounds);
				}
			}
		}
		private void DrawLastRefresh(int index, Rectangle bounds, Graphics g, bool selected) {
			Brush b = (selected ? SystemBrushes.HighlightText : SystemBrushes.ControlDark);
			
			
			g.DrawString(Startup.FormatTime(friends[index].LastRefresh), Font, b, bounds, itemFormat);
		}


		private void LoadIcons() {
			if(icons != null && icons.Length > 0) {
				for(int i = 0; i < icons.Length; i++) {
					lock(icons[i]) {
						icons[i].Dispose();
					}
				}
			}

			lock(this) {
				icons = new Image[friends.Length];
			}

			for(int i = 0; i < friends.Length; i++) {
				lock(icons) {
					icons[i] = Startup.IconCache[friends[i].UserName];

				}
			}
		}

		protected override void Dispose(bool disposing) {
			base.Dispose (disposing);
			mailIcon.Dispose();
		}


		public int PreferedSize {
			get {
				return Items.Count * ItemHeight;
			}
		}
	}
}
